/*
 * Chron
 * Copyright Jonathan Booth, Oct 13, 1996
 *
 * Last Update: Oct 13, 1996
 *
 * A rewrite of the X-windows clock 'Chron', except prettier.
 */

#include <time.h>
#include "Clock.h"
#include "Color.h"
#include "Messages.h"

#define FaceFactor 0.4
#define SecFactor 0.45
#define MinFactor 0.3
#define HourFactor 0.2
#define delay 100000

Face::Face(BRect Area, color_space Mode)
	: BBitmap(Area,Mode,TRUE)
{
	// Get setup for creation
	BPoint center((Area.right - Area.left) / 2.0,
				  (Area.bottom - Area.top) / 2.0);
	float LenX = (Area.right - Area.left) * FaceFactor;
	float LenY = (Area.bottom - Area.top) * FaceFactor;
	float angle = 90;
	
	// Create it
	FaceImg = new BView(Area,"Face",0,0);
	
	// Add it in
	if (FaceImg) AddChild(FaceImg);
	
	if (Lock() && FaceImg) {
		// Clear the drawing
		FaceImg->SetHighColor(Background);
		FaceImg->FillRect(Area);
		
		rgb_color FaceBackground = {0,0,0,0};
		for (int x=0;x<360;x++) {
			FaceImg->SetHighColor(FaceBackground);
			FaceImg->FillArc(center,LenX,LenY,-x,1);
			
			FaceBackground.red = ((float)(x % 30))/31.0*255;
			FaceBackground.green = FaceBackground.red;
			FaceBackground.blue = FaceBackground.red;
		}
		
		// While there is another color to do
		for (int x=0;x<6;x++) {
			FaceImg->SetHighColor(FaceColors[x]);
			angle = angle - 30.0;
			FaceImg->FillArc(center,LenX,LenY,angle,30);

			angle = angle - 30.0;
			FaceImg->StrokeArc(center,LenX,LenY,angle,40);
		}
		
		//Redraw the red to make sure it's ok
		FaceImg->SetHighColor(FaceColors[0]);
		FaceImg->FillArc(center,LenX,LenY,90,-5);
		
		// Make sure we're ready
		FaceImg->Sync();
		
		// Release locking hooks!
		Unlock();
	}	
}

Face::~Face()
{
	if (Lock() && FaceImg) {
		RemoveChild(FaceImg);
		Unlock();
	}
	
	delete FaceImg;
}

#pragma mark -

Clock::Clock(BRect Area, color_space Mode)
	: BView(Area,"Clock",B_FOLLOW_ALL,B_WILL_DRAW)
{
	// Set up the images
	HandBit = new BBitmap(Area,Mode,TRUE);
	HandImg = new BView(Area,"Hands",0,0);
	face = new Face(Area,Mode);

	// Setup the variables
	Center.Set((Area.right - Area.left) / 2.0,
			   (Area.bottom - Area.top) / 2.0);
	LenX = Area.right - Area.left;
	LenY = Area.bottom - Area.top;
	SecStat = TRUE;
	RedrawFace = TRUE;

	// Set up the past time
	OldSec=0;
}

Clock::~Clock()
{
	if (HandBit->Lock() && HandImg) {
		HandBit->RemoveChild(HandImg);
		HandBit->Unlock();
	}
	
	delete HandImg;
}

void Clock::AttachedToWindow()
{
	BView::AttachedToWindow();
	
	// Pack drawing image into the bitmap
	if (HandImg) HandBit->AddChild(HandImg);
}

void Clock::Draw(BRect updateRect)
{
	// Draw the image the *fast* way --
	//  draw it off of a bitmap, this causes
	//  zero flicker of the image. A very good thing.
	if (Window()->Lock()) {
		DrawBitmap(HandBit,updateRect,updateRect);
		Window()->Unlock();
	}
}

void Clock::UpdateImage()
{
	// Lotsa local vars here
	time_t t;
	tm *curTime;
	float SecArc,MinArc,HourArc;
	bool isam = TRUE;
	
	// Find out the time
	time(&t);
	curTime = localtime(&t);

	/*
	 * Have the seconds changed?
	 * This has a problem in theory if the bebox
	 * hung for a minute then resumed this task,
	 * however there is a bigger problem than that
	 * if the bebox does.
	 */
	if (((int)curTime->tm_sec) != OldSec) {
		// Update the clock updated time
		OldSec = curTime->tm_sec;
		
		// Ok, update the image if it exists
		if (HandBit->Lock()) {
			/*
			 * Calculate the arc-position of the
			 * various hands
			 */
			SecArc = (float)((curTime->tm_sec + 1) * 6);
			MinArc = (float)((curTime->tm_min * 6)
				 	  + (curTime->tm_sec / 10));
			HourArc = (float)(((curTime->tm_hour % 12) * 30)
					   + (curTime->tm_min / 2.0));
		
			/*
			 * Figure out if face image needs to
			 * redraw (pm = circle shrinking) or
			 * if we just past the hour and the
			 * circles dissapeared (am)
			 */
			if (curTime->tm_hour >=12) {
				isam = FALSE;
				RedrawFace = TRUE;
			}
			if (curTime->tm_sec == 0) {
				RedrawFace = TRUE;
			}
		
			/*
			 * Did we flag a clock face redraw?
			 */
			if (RedrawFace && face) {
				RedrawFace = FALSE;
				HandImg->DrawBitmap(face,Bounds());
			}
		
			// Second hand drawing
			if (SecStat) {
				HandImg->SetHighColor(SecHandColor);
				
				// Growing or shrinking circle
				if (isam) {
					HandImg->StrokeArc(Center,LenX*SecFactor,
								   LenY*SecFactor,
								   90,-SecArc);
				} else {
					HandImg->StrokeArc(Center,LenX*SecFactor,
								   LenY*SecFactor,
								   90,364-SecArc);
				}
			}
		
			// Minute hand
			HandImg->SetHighColor(MinHandColor);
			if (isam) {
				HandImg->FillArc(Center,LenX*MinFactor,
							   LenY*MinFactor,
							   90,0-MinArc);
			} else {
				HandImg->FillArc(Center,LenX*MinFactor,
							   LenY*MinFactor,
							   90,364-MinArc);
			}

			// Hour hand
			HandImg->SetHighColor(HourHandColor);
			if (isam) {
				HandImg->FillArc(Center,LenX*HourFactor,
							   LenY*HourFactor,
							   90,0-HourArc);
			} else {
				HandImg->FillArc(Center,LenX*HourFactor,
							   LenY*HourFactor,
							   90,364-HourArc);
			}

			/*
			 * Assure that the last drawing op
			 * completed before we redraw the
			 * image
			 */
			HandImg->Sync();
		
			/* 
			 * Must unlock what we locked
			 */
			HandBit->Unlock();
		}
		
		/*
		 * Force an update of the window
		 */
		if (Window() && Window()->Lock()) {
			Draw(Window()->Bounds());
			Window()->Unlock();
		}
	}
}

void Clock::ChangeSec()
{
	/*
	 * Set the seconds status to whatever we
	 * just registered as a change and flag
	 * a redraw of the background
	 */
	SecStat = !SecStat;
	RedrawFace = TRUE;
}

#pragma mark -

ClockWindow::ClockWindow(BRect Area)
	: Window(Area,"Chron 1.1",B_TITLED_WINDOW,NULL)
{
	// setup vars
	flag = FALSE;
	
	// Get current screen mode
	screen_info cur;
	get_screen_info(&cur);
	
	// Set up stuff
	image = new Clock(Bounds(),cur.mode);
		
	// Put it in
	AddChild(image);
	Show(); 
	
	// Turn off error checking
	SetDiscipline(FALSE);

	// Set the size min & max
	SetSizeLimits(110,1280,110,1024);
	
	// Create the semaphore
	lock = create_sem(1,"Update Locker");
	
	// Start the fun
	timer = spawn_thread(Update,
						 "Update View",
						 B_NORMAL_PRIORITY,
						 this);
	resume_thread(timer);
}

ClockWindow::~ClockWindow()
{
	Window::~Window();
}

bool ClockWindow::QuitRequested()
{
	if (acquire_sem(lock) == B_NO_ERROR) {
		delete_sem(lock);
		Window::QuitRequested();
		return TRUE;
	} else {
		return FALSE;
	}
}

void ClockWindow::ChangeSec()
{
	if (image) {
		image->ChangeSec();
	}
}

long ClockWindow::Update(void *ptr)
{
	ClockWindow *obj = (ClockWindow *)ptr;
	return (obj->Update());
}

long ClockWindow::Update()
{
	while (acquire_sem(lock) == B_NO_ERROR) {
		if (!flag) {
			if (Lock() && image) {
				// Draw image
				image->UpdateImage();

				// Unlock window
				Unlock();
			}
		} else {
			Lock();

			if (image) {
				RemoveChild(image);
				delete image;
			}

			// Get current screen mode
			screen_info cur;
			get_screen_info(&cur);

			// Set up stuff
			image = new Clock(Bounds(),cur.mode);

			AddChild(image);

			flag = FALSE;
			
			Window::FrameResized(Bounds().right - Bounds().left,
								 Bounds().bottom - Bounds().top);

			Unlock();
		}
		
		release_sem(lock);
		snooze(delay);
	}
	
	return B_NO_ERROR;
}

void ClockWindow::FrameResized(float new_w,float new_h)
{
	flag = TRUE;
}